import re
import warnings

from numbers_parser.exceptions import UnsupportedWarning
from numbers_parser.generated import TSCEArchives_pb2 as TSCEArchives
from numbers_parser.functionmap import FUNCTION_MAP

from datetime import datetime, timedelta


class Formula(list):
    def __init__(self, model, row_num, col_num):
        self._stack = []
        self._model = model
        self.row = row_num
        self.col = col_num

    def __str__(self):
        return "".join(reversed(self._stack))

    def pop(self) -> str:
        return self._stack.pop()

    def popn(self, num_args: int) -> tuple:
        values = ()
        for i in range(num_args):
            values += (self._stack.pop(),)
        return values

    def push(self, val: str):
        self._stack.append(val)

    def add(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}+{arg2}")

    def array(self, *args):
        node = args[2]
        num_rows = node.AST_array_node_numRow
        num_cols = node.AST_array_node_numCol
        if num_rows == 1:
            # 1-dimentional array: {a,b,c,d}
            args = self.popn(num_cols)
            args = ",".join(reversed(args))
            self.push(f"{{{args}}}")
        else:
            # 2-dimentional array: {a,b;c,d}
            rows = []
            for row_num in range(num_rows):
                args = self.popn(num_cols)
                args = ",".join(reversed(args))
                rows.append(f"{args}")
            args = ";".join(reversed(rows))
            self.push(f"{{{args}}}")

    def boolean(self, *args):
        node = args[2]
        if node.HasField("AST_token_node_boolean"):
            self.push(str(node.AST_token_node_boolean).upper())
        else:
            self.push(str(node.AST_boolean_node_boolean).upper())

    def concat(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}&{arg2}")

    def date(self, *args):
        # Date literals exported as DATE()
        node = args[2]
        dt = datetime(2001, 1, 1) + timedelta(seconds=node.AST_date_node_dateNum)
        self.push(f"DATE({dt.year},{dt.month},{dt.day})")

    def div(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}÷{arg2}")

    def empty(self, *args):
        self.push("")

    def equals(self, *args):
        # Arguments appear to be reversed
        arg1, arg2 = self.popn(2)
        self.push(f"{arg2}={arg1}")

    def function(self, *args):
        node = args[2]
        num_args = node.AST_function_node_numArgs
        node_index = node.AST_function_node_index
        if node_index not in FUNCTION_MAP:
            table_name = self._model.table_name(self._table_id)
            warnings.warn(
                f"{table_name}@[{self.row},{self.col}]: function ID {node_index} is unsupported",
                UnsupportedWarning,
            )
            func_name = "UNDEFINED!"
        else:
            func_name = FUNCTION_MAP[node_index]

        if len(self._stack) < num_args:
            table_name = self._model.table_name(self._table_id)
            warnings.warn(
                f"{table_name}@[{self.row},{self.col}]: stack to small for {func_name}",
                UnsupportedWarning,
            )
            num_args = len(self._stack)

        args = self.popn(num_args)
        args = ",".join(reversed(args))
        self.push(f"{func_name}({args})")

    def greater_than(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}>{arg2}")

    def greater_than_or_equal(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}≥{arg2}")

    def less_than(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}<{arg2}")

    def less_than_or_equal(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}≤{arg2}")

    def list(self, *args):
        node = args[2]
        args = self.popn(node.AST_list_node_numArgs)
        args = ",".join(reversed(args))
        self.push(f"({args})")

    def mul(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}×{arg2}")

    def negate(self, *args):
        arg1 = self.pop()
        self.push(f"-{arg1}")

    def not_equals(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}≠{arg2}")

    def number(self, *args):
        node = args[2]
        if node.AST_number_node_decimal_high == 0x3040000000000000:
            # Integer: don't use decimals
            self.push(str(node.AST_number_node_decimal_low))
        else:
            # TODO: detect when scientific notation is present
            self.push(number_to_str(node.AST_number_node_number))

    def percent(self, *args):
        arg1 = self.pop()
        self.push(f"{arg1}%")

    def power(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}^{arg2}")

    def range(self, *args):
        arg2, arg1 = self.popn(2)
        func_range = "(" in arg1 or "(" in arg2
        if "::" in arg1 and not func_range:
            # Assumes references are not cross-table
            arg1_parts = arg1.split("::")
            arg2_parts = arg2.split("::")
            self.push(f"{arg1_parts[0]}::{arg1_parts[1]}:{arg2_parts[1]}")
        else:
            self.push(f"{arg1}:{arg2}")

    def string(self, *args):
        node = args[2]
        self.push('"' + node.AST_string_node_string + '"')

    def sub(self, *args):
        arg2, arg1 = self.popn(2)
        self.push(f"{arg1}-{arg2}")

    def unsupported(self, *args):
        pass

    def xref(self, *args):
        (row_num, col_num, node) = args
        self.push(self._model.node_to_ref(row_num, col_num, node))


NODE_FUNCTION_MAP = {
    "ADDITION_NODE": "add",
    "APPEND_WHITESPACE_NODE": None,
    "ARRAY_NODE": "array",
    "BEGIN_EMBEDDED_NODE_ARRAY": None,
    "BOOLEAN_NODE": "boolean",
    "CELL_REFERENCE_NODE": "xref",
    "COLON_NODE": "range",
    # "COLON_NODE_WITH_UIDS": "unsupported",
    "CONCATENATION_NODE": "concat",
    "DATE_NODE": "date",
    "DIVISION_NODE": "div",
    "END_THUNK_NODE": None,
    "EMPTY_ARGUMENT_NODE": "empty",
    "EQUAL_TO_NODE": "equals",
    "FUNCTION_NODE": "function",
    "GREATER_THAN_NODE": "greater_than",
    "GREATER_THAN_OR_EQUAL_TO_NODE": "greater_than_or_equal",
    "LESS_THAN_NODE": "less_than",
    "LESS_THAN_OR_EQUAL_TO_NODE": "less_than_or_equal",
    "LIST_NODE": "list",
    "MULTIPLICATION_NODE": "mul",
    "NEGATION_NODE": "negate",
    "NOT_EQUAL_TO_NODE": "not_equals",
    "NUMBER_NODE": "number",
    "PERCENT_NODE": "percent",
    "POWER_NODE": "power",
    "PREPEND_WHITESPACE_NODE": None,
    "STRING_NODE": "string",
    "SUBTRACTION_NODE": "sub",
    "TOKEN_NODE": "boolean",
}


class TableFormulas:
    def __init__(self, model, table_id):
        self._model = model
        self._table_id = table_id
        self._formula_type_lookup = {
            k: v.name
            for k, v in TSCEArchives._ASTNODEARRAYARCHIVE_ASTNODETYPE.values_by_number.items()
        }

    def is_formula(self, row, col):
        return (row, col) in self._model.formula_cell_ranges(self._table_id)

    def formula(self, formula_key, row_num, col_num):
        all_formulas = self._model.formula_ast(self._table_id)
        if formula_key not in all_formulas:
            table_name = self._model.table_name(self._table_id)
            warnings.warn(
                f"{table_name}@[{row_num},{col_num}]: key #{formula_key} not found",
                UnsupportedWarning,
            )
            return "INVALID_KEY!(" + str(formula_key) + ")"

        formula = Formula(self._model, row_num, col_num)
        for node in all_formulas[formula_key]:
            node_type = self._formula_type_lookup[node.AST_node_type]
            if node_type not in NODE_FUNCTION_MAP:
                table_name = self._model.table_name(self._table_id)
                warnings.warn(
                    f"{table_name}@[{row_num},{col_num}]: node type {node_type} is unsupported",
                    UnsupportedWarning,
                )
                pass
            elif NODE_FUNCTION_MAP[node_type] is not None:
                func = getattr(formula, NODE_FUNCTION_MAP[node_type])
                func(row_num, col_num, node)

        return str(formula)


def number_to_str(v: int) -> str:
    v_str = repr(v)
    if "e" in v_str:
        number, exp = v_str.split("e")
        number = re.sub(r"[,-.]", "", number)
        zeroes = "0" * (abs(int(exp)) - 1)
        if v < 0:
            sign = "-"
        else:
            sign = ""
        if int(exp) > 0:
            return f"{sign}{number}{zeroes}"
        else:
            return f"{sign}0.{zeroes}{number}"
    else:
        return v_str



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: TNCommandArchives_sos.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import numbers_parser.generated.TSACommandArchives_sos_pb2 as TSACommandArchives__sos__pb2
import numbers_parser.generated.TSKArchives_pb2 as TSKArchives__pb2
import numbers_parser.generated.TSPMessages_pb2 as TSPMessages__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='TNCommandArchives_sos.proto',
  package='TNSOS',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1bTNCommandArchives_sos.proto\x12\x05TNSOS\x1a\x1cTSACommandArchives_sos.proto\x1a\x11TSKArchives.proto\x1a\x11TSPMessages.proto\"\x96\x01\n-InducedVerifyDocumentWithServerCommandArchive\x12\"\n\x05super\x18\x01 \x02(\x0b\x32\x13.TSK.CommandArchive\x12 \n\rsheet_id_list\x18\x02 \x03(\x0b\x32\t.TSP.UUID\x12\x1f\n\x17sheet_id_list_undefined\x18\x03 \x01(\x08\"\x83\x01\n4InducedVerifyDrawableZOrdersWithServerCommandArchive\x12K\n\x05super\x18\x01 \x02(\x0b\x32<.TSASOS.InducedVerifyDrawableZOrdersWithServerCommandArchive'
  ,
  dependencies=[TSACommandArchives__sos__pb2.DESCRIPTOR,TSKArchives__pb2.DESCRIPTOR,TSPMessages__pb2.DESCRIPTOR,])




_INDUCEDVERIFYDOCUMENTWITHSERVERCOMMANDARCHIVE = _descriptor.Descriptor(
  name='InducedVerifyDocumentWithServerCommandArchive',
  full_name='TNSOS.InducedVerifyDocumentWithServerCommandArchive',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='super', full_name='TNSOS.InducedVerifyDocumentWithServerCommandArchive.super', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sheet_id_list', full_name='TNSOS.InducedVerifyDocumentWithServerCommandArchive.sheet_id_list', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sheet_id_list_undefined', full_name='TNSOS.InducedVerifyDocumentWithServerCommandArchive.sheet_id_list_undefined', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=107,
  serialized_end=257,
)


_INDUCEDVERIFYDRAWABLEZORDERSWITHSERVERCOMMANDARCHIVE = _descriptor.Descriptor(
  name='InducedVerifyDrawableZOrdersWithServerCommandArchive',
  full_name='TNSOS.InducedVerifyDrawableZOrdersWithServerCommandArchive',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='super', full_name='TNSOS.InducedVerifyDrawableZOrdersWithServerCommandArchive.super', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=260,
  serialized_end=391,
)

_INDUCEDVERIFYDOCUMENTWITHSERVERCOMMANDARCHIVE.fields_by_name['super'].message_type = TSKArchives__pb2._COMMANDARCHIVE
_INDUCEDVERIFYDOCUMENTWITHSERVERCOMMANDARCHIVE.fields_by_name['sheet_id_list'].message_type = TSPMessages__pb2._UUID
_INDUCEDVERIFYDRAWABLEZORDERSWITHSERVERCOMMANDARCHIVE.fields_by_name['super'].message_type = TSACommandArchives__sos__pb2._INDUCEDVERIFYDRAWABLEZORDERSWITHSERVERCOMMANDARCHIVE
DESCRIPTOR.message_types_by_name['InducedVerifyDocumentWithServerCommandArchive'] = _INDUCEDVERIFYDOCUMENTWITHSERVERCOMMANDARCHIVE
DESCRIPTOR.message_types_by_name['InducedVerifyDrawableZOrdersWithServerCommandArchive'] = _INDUCEDVERIFYDRAWABLEZORDERSWITHSERVERCOMMANDARCHIVE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

InducedVerifyDocumentWithServerCommandArchive = _reflection.GeneratedProtocolMessageType('InducedVerifyDocumentWithServerCommandArchive', (_message.Message,), {
  'DESCRIPTOR' : _INDUCEDVERIFYDOCUMENTWITHSERVERCOMMANDARCHIVE,
  '__module__' : 'TNCommandArchives_sos_pb2'
  # @@protoc_insertion_point(class_scope:TNSOS.InducedVerifyDocumentWithServerCommandArchive)
  })
_sym_db.RegisterMessage(InducedVerifyDocumentWithServerCommandArchive)

InducedVerifyDrawableZOrdersWithServerCommandArchive = _reflection.GeneratedProtocolMessageType('InducedVerifyDrawableZOrdersWithServerCommandArchive', (_message.Message,), {
  'DESCRIPTOR' : _INDUCEDVERIFYDRAWABLEZORDERSWITHSERVERCOMMANDARCHIVE,
  '__module__' : 'TNCommandArchives_sos_pb2'
  # @@protoc_insertion_point(class_scope:TNSOS.InducedVerifyDrawableZOrdersWithServerCommandArchive)
  })
_sym_db.RegisterMessage(InducedVerifyDrawableZOrdersWithServerCommandArchive)


# @@protoc_insertion_point(module_scope)



class NumbersError(Exception):
    """Base class for other exceptions"""

    pass


class UnsupportedError(NumbersError):
    """Raised for unsupported file format features"""

    pass


class NotImplementedError(NumbersError):
    """Raised for missing Protobufs"""

    pass


class FileError(NumbersError):
    """Raised for IO and other OS errors"""

    pass


class FileFormatError(NumbersError):
    """Raised for parsing errors during file load"""

    pass


class FormulaError(NumbersError):
    """ "Raise for formula evaluation errors"""

    pass


class UnsupportedWarning(Warning):
    """Raised for unsupported file format features"""

    pass



import re

from numbers_parser.generated import TSTArchives_pb2 as TSTArchives
from numbers_parser.exceptions import UnsupportedError

from datetime import timedelta
from logging import debug


class Cell:
    @staticmethod
    def factory(model: object, table_id: int, row_num: int, col_num: int):  # NOQA: C901
        data = model.table_cell_decode(table_id, row_num, col_num)

        row_col = (row_num, col_num)
        merge_cells = model.merge_cell_ranges(table_id)
        is_merged = row_col in merge_cells

        if data is None:
            if is_merged and merge_cells[row_col]["merge_type"] == "ref":
                cell = MergedCell(*merge_cells[row_col]["rect"])
            else:
                cell = EmptyCell(row_num, col_num, None)
            cell.size = None
            cell._model = model
            cell._table_id = table_id
            return cell

        cell_type = data["type"]
        if cell_type == TSTArchives.emptyCellValueType:
            cell = EmptyCell(row_num, col_num, None)
        elif cell_type == TSTArchives.numberCellType:
            cell = NumberCell(row_num, col_num, data["value"])
        elif cell_type == TSTArchives.textCellType:
            cell = TextCell(row_num, col_num, data["value"])
        elif cell_type == TSTArchives.dateCellType:
            cell = DateCell(row_num, col_num, data["value"])
        elif cell_type == TSTArchives.boolCellType:
            cell = BoolCell(row_num, col_num, data["value"])
        elif cell_type == TSTArchives.durationCellType:
            cell = DurationCell(row_num, col_num, timedelta(days=data["value"]))
        elif cell_type == TSTArchives.formulaErrorCellType:
            cell = ErrorCell(row_num, col_num, None)
        elif cell_type == TSTArchives.currencyCellValueType:
            cell = FormulaCell(row_num, col_num, data["value"])
        else:
            raise UnsupportedError(  # pragma: no cover
                f"Unsupport cell type {cell_type} @:({row_num},{col_num})"
            )

        cell._table_id = table_id
        cell._model = model

        if is_merged and merge_cells[row_col]["merge_type"] == "source":
            cell.is_merged = True
            cell.size = merge_cells[row_col]["size"]

        debug(
            "%s@[%d,%d]: value=%s",
            model.table_name(table_id),
            row_num,
            col_num,
            str(cell.value),
        )
        return cell

    def __init__(self, row_num: int, col_num: int, value):
        self._value = value
        self.row = row_num
        self.col = col_num
        self.size = (1, 1)
        self.is_merged = False

    @property
    def formula(self):
        if hasattr(self, "_formula"):
            return self._formula

        formula_key = self._model.table_cell_formula_decode(
            self._table_id,
            self.row,
            self.col,
            self._type,
        )
        if formula_key is not None:
            try:
                table_formulas = self._model.table_formulas(self._table_id)
                formula = table_formulas.formula(formula_key, self.row, self.col)
            except KeyError:  # pragma: no cover
                raise UnsupportedError(f"Formula not found ({self.row},{self.col})")
            except IndexError:  # pragma: no cover
                raise UnsupportedError(
                    f"Unsupported formula buffer ({self.row},{self.col})"
                )
            self._formula = formula

            debug(
                "%s@[%d,%d]: key=%d:%s: type=%d, value=%s",
                self._model.table_name(self._table_id),
                self.row,
                self.col,
                formula_key,
                formula,
                self._type,
                str(self.value),
            )
        else:
            self._formula = None

        return self._formula


class NumberCell(Cell):
    def __init__(self, row_num: int, col_num: int, value):
        self._type = TSTArchives.numberCellType
        super().__init__(row_num, col_num, value)

    @property
    def value(self) -> int:
        return self._value


class TextCell(Cell):
    def __init__(self, row_num: int, col_num: int, value):
        self._type = TSTArchives.textCellType
        super().__init__(row_num, col_num, value)

    @property
    def value(self) -> str:
        return self._value


class EmptyCell(Cell):
    def __init__(self, row_num: int, col_num: int, value):
        self._type = None
        super().__init__(row_num, col_num, value)

    @property
    def value(self):
        return None


class BoolCell(Cell):
    def __init__(self, row_num: int, col_num: int, value):
        self._type = TSTArchives.boolCellType
        super().__init__(row_num, col_num, value)

    @property
    def value(self) -> bool:
        return self._value


class DateCell(Cell):
    def __init__(self, row_num: int, col_num: int, value):
        self._type = TSTArchives.dateCellType
        super().__init__(row_num, col_num, value)

    @property
    def value(self) -> timedelta:
        return self._value


class DurationCell(Cell):
    def __init__(self, row_num: int, col_num: int, value):
        self._type = TSTArchives.durationCellType
        super().__init__(row_num, col_num, value)

    @property
    def value(self) -> timedelta:
        return self._value


class FormulaCell(Cell):
    def __init__(self, row_num: int, col_num: int, value):
        self._type = TSTArchives.formulaCellType
        super().__init__(row_num, col_num, value)

    @property
    def value(self):
        return None


class ErrorCell(Cell):
    def __init__(self, row_num: int, col_num: int, value):
        self._type = TSTArchives.formulaErrorCellType
        super().__init__(row_num, col_num, value)

    @property
    def value(self):
        return None


class MergedCell:
    def __init__(self, row_start: int, col_start: int, row_end: int, col_end: int):
        self.value = None
        self.formula = None
        self.row_start = row_start
        self.row_end = row_end
        self.col_start = col_start
        self.col_end = col_end
        self.merge_range = xl_range(row_start, col_start, row_end, col_end)


# Cell reference conversion from  https://github.com/jmcnamara/XlsxWriter
# Copyright (c) 2013-2021, John McNamara <jmcnamara@cpan.org>
range_parts = re.compile(r"(\$?)([A-Z]{1,3})(\$?)(\d+)")


def xl_cell_to_rowcol(cell_str: str) -> tuple:
    """
    Convert a cell reference in A1 notation to a zero indexed row and column.
    Args:
        cell_str:  A1 style string.
    Returns:
        row, col: Zero indexed cell row and column indices.
    """
    if not cell_str:
        return 0, 0

    match = range_parts.match(cell_str)
    if not match:
        raise IndexError(f"invalid cell reference {cell_str}")

    col_str = match.group(2)
    row_str = match.group(4)

    # Convert base26 column string to number.
    expn = 0
    col = 0
    for char in reversed(col_str):
        col += (ord(char) - ord("A") + 1) * (26 ** expn)
        expn += 1

    # Convert 1-index to zero-index
    row = int(row_str) - 1
    col -= 1

    return row, col


def xl_range(first_row, first_col, last_row, last_col):
    """
    Convert zero indexed row and col cell references to a A1:B1 range string.
    Args:
       first_row: The first cell row.    Int.
       first_col: The first cell column. Int.
       last_row:  The last cell row.     Int.
       last_col:  The last cell column.  Int.
    Returns:
        A1:B1 style range string.
    """
    range1 = xl_rowcol_to_cell(first_row, first_col)
    range2 = xl_rowcol_to_cell(last_row, last_col)

    if range1 == range2:
        return range1
    else:
        return range1 + ":" + range2


def xl_rowcol_to_cell(row, col, row_abs=False, col_abs=False):
    """
    Convert a zero indexed row and column cell reference to a A1 style string.
    Args:
       row:     The cell row.    Int.
       col:     The cell column. Int.
       row_abs: Optional flag to make the row absolute.    Bool.
       col_abs: Optional flag to make the column absolute. Bool.
    Returns:
        A1 style string.
    """
    if row < 0:
        raise IndexError(f"Row reference {row} below zero")

    if col < 0:
        raise IndexError(f"column reference {col} below zero")

    row += 1  # Change to 1-index.
    row_abs = "$" if row_abs else ""

    col_str = xl_col_to_name(col, col_abs)

    return col_str + row_abs + str(row)


def xl_col_to_name(col, col_abs=False):
    """
    Convert a zero indexed column cell reference to a string.
    Args:
       col:     The cell column. Int.
       col_abs: Optional flag to make the column absolute. Bool.
    Returns:
        Column style string.
    """
    col_num = col
    if col_num < 0:
        raise IndexError(f"Column reference {col_num} below zero")

    col_num += 1  # Change to 1-index.
    col_str = ""
    col_abs = "$" if col_abs else ""

    while col_num:
        # Set remainder from 1 .. 26
        remainder = col_num % 26

        if remainder == 0:
            remainder = 26

        # Convert the remainder to a character.
        col_letter = chr(ord("A") + remainder - 1)

        # Accumulate the column letters, right to left.
        col_str = col_letter + col_str

        # Get the next order of magnitude.
        col_num = int((col_num - 1) / 26)

    return col_abs + col_str



from numbers_parser.unpack import read_numbers_file


class ItemsList:
    def __init__(self, model, refs, item_class):
        self._item_name = item_class.__name__.lower()
        self._items = [item_class(model, _) for _ in refs]

    def __getitem__(self, key: int):
        if type(key) == int:
            if key < 0 or key >= len(self._items):
                raise IndexError(f"index {key} out of range")
            return self._items[key]
        elif type(key) == str:
            for item in self._items:
                if item.name == key:
                    return item
            raise KeyError(f"no {self._item_name} named '{key}'")
        else:
            t = type(key).__name__
            raise LookupError(f"invalid index type {t}")

    def __len__(self) -> int:
        return len(self._items)


class ObjectStore:
    def __init__(self, path: str) -> int:
        self._objects = {}
        read_numbers_file(
            path,
            handler=lambda identifier, obj: self._store_object(identifier, obj),
            store_objects=True,
        )

    def _store_object(self, identifier, obj):
        self._objects[identifier] = obj

    def __getitem__(self, key: str):
        return self._objects[key]

    def __len__(self) -> int:
        return len(self._objects)

    def find_refs(self, ref_name) -> list:
        refs = [k for k, v in self._objects.items() if type(v).__name__ == ref_name]
        return refs



from setuptools import setup, find_packages
from distutils.util import convert_path

main_ns = {}
ver_path = convert_path("src/numbers_parser/_version.py")
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

setup(
    name="numbers-parser",
    version=main_ns["__version__"],
    author="Jon Connell",
    author_email="python@figsandfudge.com",
    description="Package to read data from Apple Numbers spreadsheets",
    license="MIT",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/masaccio/numbers-parser",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    entry_points={
        "console_scripts": [
            "cat-numbers=numbers_parser._cat_numbers:main",
            "unpack-numbers=numbers_parser._unpack_numbers:main",
        ],
    },
    install_requires=[
        "protobuf",
        "pytest",
        "python-snappy",
        "PyYAML",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



# Forked from https://github.com/psobot/keynote-parser/blob/master/keynote_parser/codec.py

import struct
import snappy
from functools import partial

from numbers_parser.mapping import ID_NAME_MAP
from numbers_parser.exceptions import NotImplementedError

from google.protobuf.internal.decoder import _DecodeVarint32
from google.protobuf.json_format import MessageToDict

from numbers_parser.generated.TSPArchiveMessages_pb2 import ArchiveInfo


class IWAFile(object):
    def __init__(self, chunks, filename=None):
        self.chunks = chunks
        self.filename = filename

    @classmethod
    def from_buffer(cls, data, filename=None):
        try:
            chunks = []
            while data:
                chunk, data = IWACompressedChunk.from_buffer(data, filename)
                chunks.append(chunk)

            return cls(chunks, filename)
        except Exception as e:  # pragma: no cover
            if filename:
                raise ValueError("Failed to deserialize " + filename) from e
            else:
                raise

    def to_dict(self):
        try:
            return {"chunks": [chunk.to_dict() for chunk in self.chunks]}
        except Exception as e:  # pragma: no cover
            if self.filename:
                raise ValueError("Failed to serialize " + self.filename) from e
            else:
                raise


class IWACompressedChunk(object):
    def __init__(self, archives):
        self.archives = archives

    def __eq__(self, other):
        return self.archives == other.archives  # pragma: no cover

    @classmethod
    def _decompress_all(cls, data):
        while data:
            header = data[:4]

            first_byte = header[0]
            if not isinstance(first_byte, int):
                first_byte = ord(first_byte)

            if first_byte != 0x00:
                raise ValueError(  # pragma: no cover
                    "IWA chunk does not start with 0x00! (found %x)" % first_byte
                )

            unpacked = struct.unpack_from("<I", bytes(header[1:]) + b"\x00")
            length = unpacked[0]
            chunk = data[4 : 4 + length]
            data = data[4 + length :]

            try:
                yield snappy.uncompress(chunk)
            except Exception:  # pragma: no cover
                # Try to see if this data isn't compressed in the first place.
                # If this data is still compressed, parsing it as Protobuf
                # will almost definitely fail anyways.
                yield chunk

    @classmethod
    def from_buffer(cls, data, filename=None):
        data = b"".join(cls._decompress_all(data))
        archives = []
        while data:
            archive, data = IWAArchiveSegment.from_buffer(data, filename)
            archives.append(archive)
        return cls(archives), None

    def to_dict(self):
        return {"archives": [archive.to_dict() for archive in self.archives]}


class ProtobufPatch(object):
    def __init__(self, data):
        self.data = data

    def __eq__(self, other):
        return self.data == other.data  # pragma: no cover

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.data)  # pragma: no cover

    def to_dict(self):
        return message_to_dict(self.data)

    @classmethod
    def FromString(cls, message_info, proto_klass, data):
        # Recent numbers does not apear to store date this way
        assert len(message_info.diff_field_path.path) != 1
        return cls(proto_klass.FromString(data))


class IWAArchiveSegment(object):
    def __init__(self, header, objects):
        self.header = header
        self.objects = objects

    def __eq__(self, other):
        return (
            self.header == other.header and self.objects == other.objects
        )  # pragma: no cover

    def __repr__(self):
        return "<%s identifier=%s objects=%s>" % (  # pragma: no cover
            self.__class__.__name__,
            self.header.identifier,
            repr(self.objects).replace("\n", " ").replace("  ", " "),
        )

    @classmethod
    def from_buffer(cls, buf, filename=None):
        archive_info, payload = get_archive_info_and_remainder(buf)
        if not repr(archive_info):
            raise ValueError(
                "Segment doesn't seem to start with an ArchiveInfo!"
            )  # pragma: no cover

        payloads = []

        n = 0
        for message_info in archive_info.message_infos:
            try:
                if message_info.type == 0 and archive_info.should_merge and payloads:
                    base_message = archive_info.message_infos[
                        message_info.base_message_index
                    ]
                    klass = partial(
                        ProtobufPatch.FromString,
                        message_info,
                        ID_NAME_MAP[base_message.type],
                    )
                else:
                    klass = ID_NAME_MAP[message_info.type]
            except KeyError:  # pragma: no cover
                raise NotImplementedError(
                    "Don't know how to parse Protobuf message type "
                    + str(message_info.type)
                )
            try:
                message_payload = payload[n : n + message_info.length]
                if hasattr(klass, "FromString"):
                    output = klass.FromString(message_payload)
                else:
                    output = klass(message_payload)
            except Exception as e:  # pragma: no cover
                raise ValueError(
                    "Failed to deserialize %s payload of length %d: %s"
                    % (klass, message_info.length, e)
                )
            payloads.append(output)
            n += message_info.length

        return cls(archive_info, payloads), payload[n:]

    def to_dict(self):
        return {
            "header": header_to_dict(self.header),
            "objects": [message_to_dict(message) for message in self.objects],
        }


def message_to_dict(message):
    if hasattr(message, "to_dict"):
        return message.to_dict()
    output = MessageToDict(message)
    output["_pbtype"] = type(message).DESCRIPTOR.full_name
    return output


def header_to_dict(message):
    output = message_to_dict(message)
    for message_info in output["messageInfos"]:
        del message_info["length"]
    return output


def get_archive_info_and_remainder(buf):
    msg_len, new_pos = _DecodeVarint32(buf, 0)
    n = new_pos
    msg_buf = buf[n : n + msg_len]
    n += msg_len
    return ArchiveInfo.FromString(msg_buf), buf[n:]
